--- This file has been generated by scripts/build_db.py. DO NOT EDIT !

INSERT INTO "vertical_datum_ensemble_member" VALUES('EPSG','1288','EPSG','5130',1);
INSERT INTO "vertical_datum_ensemble_member" VALUES('EPSG','1288','EPSG','5131',2);
INSERT INTO "vertical_datum_ensemble_member" VALUES('EPSG','1288','EPSG','5101',3);
INSERT INTO "vertical_datum_ensemble_member" VALUES('EPSG','1288','EPSG','1164',4);
INSERT INTO "vertical_datum_ensemble_member" VALUES('EPSG','1288','EPSG','5138',5);
INSERT INTO "vertical_datum_ensemble_member" VALUES('EPSG','1288','EPSG','5140',6);
INSERT INTO "vertical_datum_ensemble_member" VALUES('EPSG','1288','EPSG','5144',7);
INSERT INTO "vertical_datum_ensemble_member" VALUES('EPSG','1288','EPSG','5148',8);
INSERT INTO "vertical_datum_ensemble_member" VALUES('EPSG','1288','EPSG','5147',9);
